part of katana_routing;

extension NavigatorStateExtensions on NavigatorState {
  /// Pops the page to the page with the specified path.
  ///
  /// Enter the pathname of the page in [name].
  Future<void> popUntilNamed(String name) async {
    popUntil((route) {
      return route.settings.name == name;
    });
  }

  /// Reset all history and go to a specific page.
  ///
  /// After all pages are [pop()],
  /// the page specified by [newRouteName] is [pushNamed()].
  ///
  /// In [arguments], put [RouteQuery] and so on.
  ///
  /// Ideal for moving after logging out.
  Future<T?> resetAndPushNamed<T extends Object?>(
    String newRouteName, {
    Object? arguments,
  }) {
    return pushNamedAndRemoveUntil(
      newRouteName,
      (route) => false,
      arguments: arguments,
    );
  }
}

extension BuildContedxtExtensions on BuildContext {
  /// Outputs the theme related to the context.
  ThemeData get theme => Theme.of(this);

  /// Outputs the widget theme related to the context.
  WidgetTheme get widgetTheme => WidgetTheme.of(this);

  /// Get the Navigator related to context.
  NavigatorState get navigator => Navigator.of(this);

  /// Get the Root navigator related to context.
  NavigatorState get rootNavigator => Navigator.of(this, rootNavigator: true);

  /// Get the Flavor.
  String get flavor => FlavorScope.of(this).flavor;

  /// Get the Design type.
  DesignType get designType => DesignTypeScope.of(this).designType;

  /// If you want to use the style for web when you are on the web, use `true`.
  bool get webStyle => DesignTypeScope.of(this).webStyle;

  /// Get the route settins.
  RouteSettings get route =>
      ModalRoute.of(this)?.settings ?? const RouteSettings();

  /// Get the media qury related to context
  MediaQueryData get mediaQuery => MediaQuery.of(this);

  /// Releases focus such as text field.
  void unfocus() {
    FocusManager.instance.primaryFocus?.unfocus();
  }
}

extension WidgetRefExtensions on WidgetRef {
  /// Explicitly rebuild the widget.
  ///
  /// Basically, do not use it.
  ///
  /// Use this only if you want to force a rebuild.
  void rebuild() {
    Scoped.of(this as BuildContext).rebuild();
  }

  /// Open a new URL or screen.
  ///
  /// If [path] starts with `http`,
  /// it opens a browser, otherwise it opens a screen path.
  ///
  /// It is possible to configure the settings for
  /// opening the screen by specifying [query].
  Future<T?> open<T extends Object?>(String path, [RouteQuery? query]) async {
    if (path.startsWith("http")) {
      await openURL(path);
      return null;
    } else {
      return await (this as BuildContext)
          .navigator
          .pushNamed(path, arguments: query);
    }
  }

  /// Get the data passed to the page.
  ///
  /// Basically it returns the [RouteQuery] data passed from the previous page.
  ///
  /// If the [RouteQuery] data is not found,
  /// the map stored in [PageScopedWidget] is retrieved.
  DynamicMap get arg {
    final pageScoped = _PageScoped.of(this as BuildContext);
    return pageScoped._modalRoute?.settings.arguments as DynamicMap? ??
        pageScoped._map;
  }

  /// Gets the value from the [key] of the map stored in [arg].
  ///
  /// If [key] does not exist in the map, [orElse] is returned.
  T get<T>(String key, T orElse) {
    return arg.get(key, orElse);
  }

  /// Get the list corresponding to [key] in the map.
  ///
  /// If [key] is not found, the list of [orElse] is returned.
  List<T> getAsList<T>(String key, [List<T>? orElse]) {
    if (!arg.containsKey(key)) {
      return orElse ?? [];
    }
    return (arg[key] as List?)?.cast<T>() ?? orElse ?? [];
  }

  /// Get the map corresponding to [key] in the map.
  ///
  /// If [key] is not found, the map of [orElse] is returned.
  Map<String, T> getAsMap<T>(String key, [Map<String, T>? orElse]) {
    if (!arg.containsKey(key)) {
      return orElse ?? {};
    }
    return (arg[key] as Map?)?.cast<String, T>() ?? orElse ?? {};
  }

  /// Get the set corresponding to [key] in the map.
  ///
  /// If [key] is not found, the set of [orElse] is returned.
  Set<T> getAsSets<T>(String key, [Set<T>? orElse]) {
    if (!arg.containsKey(key)) {
      return orElse ?? {};
    }
    return (arg[key] as Set?)?.cast<T>() ?? orElse ?? {};
  }

  /// Get the set corresponding to [key] in the DateTime.
  ///
  /// If [key] is not found, the set of [orElse] is returned.
  DateTime getAsDateTime(String key, [DateTime? orElse]) {
    if (!arg.containsKey(key)) {
      return orElse ?? DateTime.now();
    }
    final millisecondsSinceEpoch = arg[key] as int?;
    if (millisecondsSinceEpoch == null) {
      return orElse ?? DateTime.now();
    }
    return DateTime.fromMillisecondsSinceEpoch(millisecondsSinceEpoch);
  }

  /// Get [value] to [key] of the map stored in [arg].
  ///
  /// The data will be unique within the page and can continue to be used even if the widget is rebuilt.
  dynamic operator [](Object? key) => arg[key];

  /// Save [value] to [key] of the map stored in [arg].
  ///
  /// The data will be unique within the page and can continue to be used even if the widget is rebuilt.
  void operator []=(String key, dynamic value) {
    arg[key] = value;
  }

  /// A builder for creating and saving containers.
  ///
  /// If [key] is specified, the data will be saved with that key.
  ///
  /// Since it is not usually used directly from [get], [prefix] is given to prevent duplication of values.
  ///
  /// It creates a container from [containerBuilder] and outputs the actual value to be used from [result] based on the container.
  ///
  /// If [test] is `true` or the container for the Key is empty, update the container.
  Result containerBuilder<Container, Result>({
    required String key,
    required Container Function(Container? prev) containerBuilder,
    required Result Function(Container container) resultBuilder,
    bool Function(Container container)? test,
    String prefix = r"_$",
  }) {
    key = "$prefix$key";
    final found = get<Container?>(key, null);
    if (found == null || (test != null && test.call(found))) {
      final container = this[key] = containerBuilder.call(found);
      if (container is VariableContainer) {
        container._ref = this;
        if (found == null) {
          container.initContainer();
        } else if (found is VariableContainer) {
          container.didUpdateContainer(found);
        }
      }
      return resultBuilder.call(container);
    }
    return resultBuilder.call(found);
  }
}

extension RoutingDynamicMapExtensions on DynamicMap {
  /// Convert the theme color from [DynamicMap].
  ThemeColor? toThemeColor() {
    return ThemeColor._fromMap(this);
  }
}

extension ThemeExtension on ThemeData {
  /// Text color.
  Color get textColor {
    return colorScheme.onBackground;
  }

  /// Text color on primary.
  Color get textColorOnPrimary {
    return colorScheme.onPrimary;
  }

  /// Text color on secondary.
  Color get textColorOnSecondary {
    return colorScheme.onSecondary;
  }

  /// Text color on surface.
  Color get textColorOnSurface {
    return colorScheme.onSurface;
  }

  /// Text color on error.
  Color get textColorOnError {
    return colorScheme.onError;
  }

  /// Secondary color.
  Color get secondaryColor {
    return colorScheme.secondary;
  }

  /// Surface color.
  Color get surfaceColor {
    return colorScheme.surface;
  }
}
